#!/usr/bin/perl

use warnings;
use strict;

use Parallel::ForkManager;
use Getopt::Long;

my ($results, $barcode_list, $indir, %index, $linker, $nprocs, $help);

$linker = 0;
$nprocs = 1;
$help = 0;
$results = GetOptions('help' => 0, 'barcodes=s' => \$barcode_list, 'linker' => \$linker, 'nprocs=i' => \$nprocs);

$indir = shift;

if (!defined($indir) || $help) {
	usage();
}

opendir(DIR, "$indir") || die("Error opening directory '$indir': $!\n");
my @files = grep { !/^\./ } readdir(DIR);
close(DIR);

my $sample_num = 1;
foreach my $barcode (split(",", $barcode_list)) {
	$barcode =~ s/^\s*//;
	$barcode =~ s/\s*$//;

	if (!($barcode =~ m/^[ACGT]+$/)) {
		print "Error: barcode '$barcode' contains non-ACGT bases.  Exiting.\n";
		exit(0);
	}

	my $sample = "Sample$sample_num";
	$sample_num++;
	$index{$sample} = {'index' => $barcode, 'barcodes' => {}, 'total' => 0};
}
$index{'Unknown'} = {'index' => 'NNNNNN', 'barcodes' => {}, 'total' => 0};

# Harcoded to split off 
my $pm = Parallel::ForkManager->new($nprocs);
foreach my $file (@files) {
	my $pid = $pm->start and next;

	my $outdir = "$indir/$file.fastq.files";

	mkdir($outdir) || die("Error making directory $outdir: $!\n");

	while (my ($index_id, $index_data) = each %index) {
		my $fh;
		open($fh, "| gzip - > $outdir/$index_id.fastq.gz") || die("Error opening file '$outdir/$index_id.fastq.gz': $!\n");
		$index_data->{'fh'} = $fh;
	}

	if ($file =~ m/\.gz/) {
		open(SEQ, "gzip -dc $file |");
	} else {
		open(SEQ, "$file");
	}
	while (!eof(SEQ)) {
		my $fastq_id = <SEQ>;
		my $seq = <SEQ>;
		my $qual_id = <SEQ>;
		my $qual = <SEQ>;
		my $barcode = substr($seq, 0, 6);
		my $min = -1;
		my @mins;

		while (my ($index_id, $index_data) = each %index) {
			next if $index_id eq 'Unknown';

			my $mm = 0;
			for (my $x = 0; $x < 6; $x++) {
				$mm++ if substr($barcode, $x, 1) ne substr($index_data->{'index'}, $x, 1);
			}

			if ($min == -1) {
				$min = $mm;
				@mins = ($index_id);
			} elsif ($mm == $min) {
				push @mins, ($index_id);
			} elsif ($mm < $min) {
				$min = $mm;
				@mins = ($index_id);
			}
		}

		# Remove the first 15 bases - just adapter sequence (only remove 9 if no linker is present
		if ($linker) {
			substr($seq, 0, 15, '');
			substr($qual, 0, 15, '');
		} else {
			substr($seq, 0, 9, '');
			substr($qual, 0, 9, '');
		}

		if (scalar(@mins) == 1 && $min < 4) {
			chomp $fastq_id;
			$fastq_id .= "($index{$mins[0]}->{'index'}|$barcode)\n";
			chomp $qual_id;
			$qual_id .= "($index{$mins[0]}->{'index'}|$barcode)\n";

			$index{$mins[0]}->{'total'}++;
			$index{$mins[0]}->{'barcodes'}->{$barcode} = 1;

			my $fh = $index{$mins[0]}->{'fh'};
			print $fh $fastq_id.$seq.$qual_id.$qual;
		} else {
			chomp $fastq_id;
			$fastq_id .= "($index{'Unknown'}->{'index'}|$barcode)\n";
			chomp $qual_id;
			$qual_id .= "($index{$mins[0]}->{'index'}|$barcode)\n";

			$index{'Unknown'}->{'total'}++;
			$index{'Unknown'}->{'barcodes'}->{$barcode} = 1;
			my $fh = $index{'Unknown'}->{'fh'};
			print $fh $fastq_id.$seq.$qual_id.$qual;
		}
	}

	my $stat_fh;
	my $stat_file = $file;
	$stat_file =~ s/\.fastq\.gz$/.split_counts.txt/;
	open($stat_fh, ">$stat_file") || die("Error opening file $stat_file: $!\n");
	print $stat_fh "index_id\tindex\ttotal_reads\ttotal_barcodes\n";
	while (my ($index_id, $index_data) = each %index) {
		print $stat_fh "$index_id\t$index_data->{'index'}\t$index_data->{'total'}\t".scalar(keys(%{$index_data->{'barcodes'}}))."\n";
	}
	close($stat_fh);

	while (my ($index_id, $index_data) = each %index) {
		my $fh = $index_data->{'fh'};
		close($fh);
	}

	$pm->finish;
}
$pm->wait_all_children;

# Merge split files
foreach my $index_id (keys(%index)) {
	my ($pid, $in_dir, @dirs, $in_fh, $out_fh);

	$pid = $pm->start and next;

	opendir($in_dir, "./");
	@dirs = grep { /fastq\.files/ } readdir($in_dir);
	close($in_dir);

	open($out_fh, "| gzip - > $index_id.fastq.gz");
	foreach my $dir (@dirs) {
		open($in_fh, "gzip -dc $dir/$index_id.fastq.gz |");
		while (my $line = <$in_fh>) {
			print $out_fh $line;
		}
		close($in_fh);
		unlink($in_fh);
	}
	close($out_fh);

	$pm->finish;
}
$pm->wait_all_children;

# Clean up split dirs
my ($in_dir, @dirs);
opendir($in_dir, "$indir");
@dirs = grep { /fastq\.files/ } readdir($in_dir);
close($in_dir);
foreach my $dir (@dirs) {
	`rm -rf $dir`;
}

sub usage {
	print "./split_barcodes.pl (options) --barcodes 'barcode1,barcode2,barcode3,etc.' <directory>\n\n";
	print "Options\n";
	print "  --help           Print this help message.\n";
	print "  --linker         Whether a linker sequence is present (default: not present)\n";
	print "  --nprocs <int>   Number of processors to use while processing the files. (default: 1)\n";

	exit(0);
}
